'''
Benchmarks for testing / comparisons
'''

import numpy as np
from freelunch.tech import ZeroLengthSolutionError

# %%

class benchmark:
    '''
    base class for benchmarking functions
    '''

    def __init__(self, n=None):
        if n is None:
            self.n = 2
        else:
            self.n = n

    def __call__(self, dna):
        if len(dna) == 0:
            raise ZeroLengthSolutionError('An empty trial solution was passed')
        return self.obj(dna)

# %%


class ackley(benchmark):
    '''
    ackely function in n dimensions
    '''

    default_bounds = lambda n:np.array([[-10, 10]]*n)
    optimum = lambda n:np.array([0]*n)
    f0 = 0
    tol = 10**-3

    a,b,c = 20, 0.2, 2*np.pi
    def obj(self, dna):
        t1 = -self.a * np.exp(-self.b * (1/len(dna)) * np.sum(dna**2))
        t2 = - np.exp(1/len(dna) * np.sum(np.cos(self.c * dna))) 
        t3 = self.a + np.exp(1)
        return t1 + t2 + t3


class exponential(benchmark):
    '''
    exponential function in n dimensions
    '''
    
    default_bounds = lambda n:np.array([[-10, 10]]*n)
    optimum = lambda n:np.array([0]*n)
    f0 = -1
    tol = 10**-3

    a = -0.5
    def obj(self, dna):
        t1 = - np.exp(self.a * np.sum(dna**2))
        return t1


class happycat(benchmark):
    '''
    happycat function in n dimensions
    '''
    
    default_bounds = lambda n:np.array([[-2, 2]]*n)
    optimum = lambda n:np.array([-1]*n)
    f0 = 0
    tol = None

    a = 1/8
    def obj(self, dna):
        norm = np.sum(dna**2)
        t1 = ((norm - self.n)**2)**(self.a)
        t2 = (1/self.n)*(0.5*norm + np.sum(dna)) + 0.5
        return t1 + t2


class periodic(benchmark):
    '''
    periodic function in n dimensions
    '''
    
    default_bounds = lambda n:np.array([[-10, 10]]*n)
    optimum = lambda n:np.array([0]*n)
    f0 = 0.9
    tol = None

    def obj(self, dna):
        t1 = np.sum(np.sin(dna)**2) +1
        t2 = - 0.1 * np.exp(np.sum(dna**2))
        return t1 + t2




"""
Main module definitions in here

"""
import numpy as np

from freelunch import tech
from freelunch.base import continuous_space_optimiser

class DE(continuous_space_optimiser):
    '''
    Differential evolution
    '''
    name = 'Differential Evolution'
    tags = ['continuous domain', 'population based', 'evolutionary']
    hyper_definitions = {
        'N':'Population size (int)',
        'G':'Number of generations (int)',
        'F':'Mutation parameter (float in [0,1])',
        'Cr':'Crossover probability (float in [0,1])'
    }
    hyper_defaults = {
        'N':100,
        'G':100,
        'F':0.5,
        'Cr':0.2
    }

    def run(self):
        pop = tech.uniform_continuous_init(self.bounds, self.hypers['N'])
        tech.compute_obj(pop, self.obj)
        for gen in range(self.hypers['G']):
            trial_pop = np.empty_like(pop, dtype=object)
            for i, sol in enumerate(pop):
                pts = np.random.choice(pop, 3)
                trial = tech.solution()
                trial.dna = (pts[0].dna - pts[1].dna) + self.hypers['F'] * pts[2].dna
                trial.dna = tech.binary_crossover(sol.dna, trial.dna, self.hypers['Cr'])
                trial.dna = tech.apply_sticky_bounds(trial.dna, self.bounds)
                trial_pop[i] = trial
            tech.compute_obj(trial_pop, self.obj)
            pop = tech.sotf(pop, trial_pop)
            jrand = np.random.randint(0, self.hypers['N'])
            pop[jrand] = trial_pop[jrand]
        return pop


class SA(continuous_space_optimiser):
    '''
    Simulated Annealing 
    '''
    name = 'Simulated Annealing'
    tags = ['Continuous domain', 'Annealing']
    hyper_definitions = {
        'K':'number of timesteps (int)',
        'T':'Temperatures (np.ndarray, T.shape=(k,))',
        'P':'Acceptance probability (P(e, e, T) -> bool)'
    }

    def run(self):
        raise NotImplementedError






'''
Standard / common techniques that are used in several optimisers are abstracted to functions here. 
'''

# %% Imports

import numpy as np


# %% Custom exceptions

class BadObjectiveFunctionScores(Exception):
    '''Exception raised when both objective score comparisons evaluate false'''

class ZeroLengthSolutionError(Exception):
    '''Exception raised when an empty solution is passed to a benchmark'''



# %% Useful classes

class solution:
    '''
    Handy dandy common object for storing trial solutions / other interesting data
    '''
    def __init__(self, dna=None, fitness=None):
        self.dna = dna
        self.fitness = None 

class adaptable_parameter:
    '''
    Class for adaptable parameters for optimisers like SADE etc.
    '''
    pass



# %% Common methods

def uniform_continuous_init(bounds, N):
    print(bounds)
    out = np.empty((N,), dtype=object)
    for i in range(N):
        out[i] = solution(np.array([np.random.uniform(a,b) for a, b in bounds]))
    return out


def compute_obj(pop, obj):
    for sol in pop:
        sol.fitness = obj(sol.dna)
        if np.isnan(sol.fitness):
            sol.fitness = None

def binary_crossover(sol1, sol2, p):
    out = np.empty_like(sol1)
    for a,b,i in zip(sol1, sol2, range(len(sol1))):
        if np.random.uniform(0,1) < p:
            out[i] = a
        else:
            out[i] = b
    return out

def sotf(olds, news):
    out = np.empty_like(olds, dtype=object)
    for old, new, i in zip(olds, news, range(len(out))):
        if new.fitness < old.fitness:
            out[i] = new
        elif old.fitness <= new.fitness:
            out[i] = old
        else: 
            raise BadObjectiveFunctionScores(
                    'Winner could not be determined by comparing objective scores. scores:{} and {}'.format(
                        old.fitness, new.fitness
                    )) 
    return out

def apply_sticky_bounds(dna, bounds):
    out = dna[:]
    for i, bound in enumerate(bounds):
        low, high = bound
        if dna[i] > high: out[i] = high
        elif dna[i] < low: out[i] = low
    return out




'''
Base classes for optimisers

'''
import numpy as np


class optimiser:
    '''
    Base class for all optimisers
    '''
    name = 'optimiser'
    tags = []
    hyper_definitions = {}
    hyper_defaults = {}
    can_run_quick = False

    def __init__(self, obj=None, hypers={}, bounds=None):
        '''
        Unlikely to instance the base class but idk what else goes here
        '''
        self.hypers = dict(self.hyper_defaults, **hypers) # Hyperparamters/ methods (dictionary of )
        self.obj = obj # Objective function
        self.bounds = bounds # Bounds / constraints
        self.nfe = 0


    def __call__(self, nruns=1, return_m=1, full_output=False):
        '''
        api for running the optimisation
        '''
        if self.obj is None:
            raise NotImplementedError('No optimiser selected')
        if nruns == 1 and return_m == 1 and not full_output and self.can_run_quick:
            return self.run_quick()
        if nruns > 1:
            sols = np.concatenate([self.run() for i in range(nruns)])
        else: 
            sols = self.run()
        sols = sorted(sols, key=lambda x: x.fitness)
        if not full_output:
            return np.array([sol.dna for sol in sols[:return_m]])
        else:
            out = {
                'optimiser':self.name,
                'hypers':self.hypers,
                'bounds':self.bounds,
                'nruns':nruns,
                'nfe':self.nfe,
                'solutions':[sol.dna for sol in sols],
                'scores':[sol.fitness for sol in sols]
            }
            return out

    def __repr__(self):
        return self.name + ' optimisation object'    

    def run(self):
        if self.obj is None:
            raise NotImplementedError('No optimiser selected')

    def run_quick(self):
        if self.obj is None:
            raise NotImplementedError('No optimiser selected')



class continuous_space_optimiser(optimiser):
    '''
    Base class for continous space optimisers i.e DE, PSO, SA etc.
    '''

class discrete_space_optimiser(optimiser):
    '''
    Base class for discrete space optimisers ie GA, GP etc.
    '''

